@@ -31,11 +31,13 @@
    Calling XPending when-ever we return to the mainloop also ensures any
    pending writes are flushed. */
 
+#include <glib.h>
 #include <stdlib.h>
 #include <limits.h>
 #include <string.h>
 #include <syslog.h>
 #include <assert.h>
+#include <unistd.h>
 #include <X11/Xatom.h>
 #include <X11/Xlib.h>
 #include <X11/extensions/Xfixes.h>
@@ -43,6 +45,10 @@
 #include "vdagent-x11.h"
 #include "vdagent-x11-priv.h"
 
+/* Stupid X11 API, there goes our encapsulate all data in a struct design */
+int (*vdagent_x11_prev_error_handler)(Display *, XErrorEvent *);
+int vdagent_x11_caught_error;
+
 static void vdagent_x11_handle_selection_notify(struct vdagent_x11 *x11,
                                                 XEvent *event, int incr);
 static void vdagent_x11_handle_selection_request(struct vdagent_x11 *x11);
@@ -68,11 +74,105 @@ static const char *vdagent_x11_sel_to_str(uint8_t selection) {
     }
 }
 
-static int debug_error_handler(Display *display, XErrorEvent *error)
+static int vdagent_x11_debug_error_handler(
+    Display *display, XErrorEvent *error)
 {
     abort();
 }
 
+/* With the clipboard we're sometimes dealing with Properties on another apps
+   Window. which can go away at any time. */
+static int vdagent_x11_ignore_bad_window_handler(
+    Display *display, XErrorEvent *error)
+{
+    if (error->error_code == BadWindow)
+        return 0;
+
+    return vdagent_x11_prev_error_handler(display, error);
+}
+
+void vdagent_x11_set_error_handler(struct vdagent_x11 *x11,
+    int (*handler)(Display *, XErrorEvent *))
+{
+    XSync(x11->display, False);
+    vdagent_x11_caught_error = 0;
+    vdagent_x11_prev_error_handler = XSetErrorHandler(handler);
+}
+
+int vdagent_x11_restore_error_handler(struct vdagent_x11 *x11)
+{
+    int error;
+
+    XSync(x11->display, False);
+    XSetErrorHandler(vdagent_x11_prev_error_handler);
+    error = vdagent_x11_caught_error;
+    vdagent_x11_caught_error = 0;
+
+    return error;
+}
+
+static void vdagent_x11_get_wm_name(struct vdagent_x11 *x11)
+{
+    Atom type_ret;
+    int format_ret;
+    unsigned long len, remain;
+    unsigned char *data = NULL;
+    Window sup_window = None;
+
+    /* XGetWindowProperty can throw a BadWindow error. One way we can trigger
+       this is when the display-manager (ie gdm) has set, and not cleared the
+       _NET_SUPPORTING_WM_CHECK property, and the window manager running in
+       the user session has not yet updated it to point to its window, so its
+       pointing to a non existing window. */
+    vdagent_x11_set_error_handler(x11, vdagent_x11_ignore_bad_window_handler);
+
+    /* Get the window manager SUPPORTING_WM_CHECK window */
+    if (XGetWindowProperty(x11->display, x11->root_window[0],
+            XInternAtom(x11->display, "_NET_SUPPORTING_WM_CHECK", False), 0,
+            LONG_MAX, False, XA_WINDOW, &type_ret, &format_ret, &len,
+            &remain, &data) == Success) {
+        if (type_ret == XA_WINDOW)
+            sup_window = *((Window *)data);
+        XFree(data);
+    }
+    if (sup_window == None &&
+        XGetWindowProperty(x11->display, x11->root_window[0],
+            XInternAtom(x11->display, "_WIN_SUPPORTING_WM_CHECK", False), 0,
+            LONG_MAX, False, XA_CARDINAL, &type_ret, &format_ret, &len,
+            &remain, &data) == Success) {
+        if (type_ret == XA_CARDINAL)
+            sup_window = *((Window *)data);
+        XFree(data);
+    }
+    /* So that we can get the net_wm_name */
+    if (sup_window != None) {
+        Atom utf8 = XInternAtom(x11->display, "UTF8_STRING", False);
+        if (XGetWindowProperty(x11->display, sup_window,
+                XInternAtom(x11->display, "_NET_WM_NAME", False), 0,
+                LONG_MAX, False, utf8, &type_ret, &format_ret, &len,
+                &remain, &data) == Success) {
+            if (type_ret == utf8) {
+                x11->net_wm_name =
+                    g_strndup((char *)data, (format_ret / 8) * len);
+            }
+            XFree(data);
+        }
+        if (x11->net_wm_name == NULL &&
+            XGetWindowProperty(x11->display, sup_window,
+                XInternAtom(x11->display, "_NET_WM_NAME", False), 0,
+                LONG_MAX, False, XA_STRING, &type_ret, &format_ret, &len,
+                &remain, &data) == Success) {
+            if (type_ret == XA_STRING) {
+                x11->net_wm_name =
+                    g_strndup((char *)data, (format_ret / 8) * len);
+            }
+            XFree(data);
+        }
+    }
+
+    vdagent_x11_restore_error_handler(x11);
+}
+
 struct vdagent_x11 *vdagent_x11_create(struct udscs_connection *vdagentd,
     int debug, int sync)
 {
@@ -96,13 +196,22 @@ struct vdagent_x11 *vdagent_x11_create(struct udscs_connection *vdagentd,
         return NULL;
     }
 
+    x11->screen_count = ScreenCount(x11->display);
+    if (x11->screen_count > MAX_SCREENS) {
+        syslog(LOG_ERR, "Error too much screens: %d > %d",
+               x11->screen_count, MAX_SCREENS);
+        XCloseDisplay(x11->display);
+        free(x11);
+        return NULL;
+    }
+
     if (sync) {
-        XSetErrorHandler(debug_error_handler);
+        XSetErrorHandler(vdagent_x11_debug_error_handler);
         XSynchronize(x11->display, True);
     }
 
-    x11->screen = DefaultScreen(x11->display);
-    x11->root_window = RootWindow(x11->display, x11->screen);
+    for (i = 0; i < x11->screen_count; i++)
+        x11->root_window[i] = RootWindow(x11->display, i);
     x11->fd = ConnectionNumber(x11->display);
     x11->clipboard_atom = XInternAtom(x11->display, "CLIPBOARD", False);
     x11->clipboard_primary_atom = XInternAtom(x11->display, "PRIMARY", False);
@@ -121,7 +230,7 @@ struct vdagent_x11 *vdagent_x11_create(struct udscs_connection *vdagentd,
     }
 
     /* We should not store properties (for selections) on the root window */
-    x11->selection_window = XCreateSimpleWindow(x11->display, x11->root_window,
+    x11->selection_window = XCreateSimpleWindow(x11->display, x11->root_window[0],
                                                 0, 0, 1, 1, 0, 0, 0);
     if (x11->debug)
         syslog(LOG_DEBUG, "Selection window: %u", (int)x11->selection_window);
@@ -131,12 +240,12 @@ struct vdagent_x11 *vdagent_x11_create(struct udscs_connection *vdagentd,
     if (XFixesQueryExtension(x11->display, &x11->xfixes_event_base, &i) &&
         XFixesQueryVersion(x11->display, &major, &minor) && major >= 1) {
         x11->has_xfixes = 1;
-        XFixesSelectSelectionInput(x11->display, x11->root_window,
+        XFixesSelectSelectionInput(x11->display, x11->root_window[0],
                                    x11->clipboard_atom,
                                    XFixesSetSelectionOwnerNotifyMask|
                                    XFixesSelectionWindowDestroyNotifyMask|
                                    XFixesSelectionClientCloseNotifyMask);
-        XFixesSelectSelectionInput(x11->display, x11->root_window,
+        XFixesSelectSelectionInput(x11->display, x11->root_window[0],
                                    x11->clipboard_primary_atom,
                                    XFixesSetSelectionOwnerNotifyMask|
                                    XFixesSelectionWindowDestroyNotifyMask|
@@ -154,14 +263,28 @@ struct vdagent_x11 *vdagent_x11_create(struct udscs_connection *vdagentd,
     if (x11->max_prop_size > 262144)
         x11->max_prop_size = 262144;
 
-    /* Catch resolution changes */
-    XSelectInput(x11->display, x11->root_window, StructureNotifyMask);
+    for (i = 0; i < x11->screen_count; i++) {
+        /* Catch resolution changes */
+        XSelectInput(x11->display, x11->root_window[i], StructureNotifyMask);
 
-    /* Get the current resolution */
-    XGetWindowAttributes(x11->display, x11->root_window, &attrib);
-    x11->width = attrib.width;
-    x11->height = attrib.height;
-    vdagent_x11_send_daemon_guest_xorg_res(x11);
+        /* Get the current resolution */
+        XGetWindowAttributes(x11->display, x11->root_window[i], &attrib);
+        x11->width[i]  = attrib.width;
+        x11->height[i] = attrib.height;
+    }
+    vdagent_x11_send_daemon_guest_xorg_res(x11, 1);
+
+    /* Get net_wm_name, since we are started at the same time as the wm,
+       sometimes we need to wait a bit for it to show up. */
+    i = 10;
+    vdagent_x11_get_wm_name(x11);
+    while (x11->net_wm_name == NULL && --i > 0) {
+        usleep(100000);
+        vdagent_x11_get_wm_name(x11);
+    }
+    if (x11->debug && x11->net_wm_name)
+        syslog(LOG_DEBUG, "net_wm_name: \"%s\", has icons: %d",
+               x11->net_wm_name, vdagent_x11_has_icons_on_desktop(x11));
 
     /* Flush output buffers and consume any pending events */
     vdagent_x11_do_read(x11);
@@ -169,18 +292,23 @@ struct vdagent_x11 *vdagent_x11_create(struct udscs_connection *vdagentd,
     return x11;
 }
 
-void vdagent_x11_destroy(struct vdagent_x11 *x11)
+void vdagent_x11_destroy(struct vdagent_x11 *x11, int vdagentd_disconnected)
 {
     uint8_t sel;
 
     if (!x11)
         return;
 
+    if (vdagentd_disconnected)
+        x11->vdagentd = NULL;
+
     for (sel = 0; sel < VD_AGENT_CLIPBOARD_SELECTION_SECONDARY; ++sel) {
         vdagent_x11_set_clipboard_owner(x11, sel, owner_none);
     }
 
     XCloseDisplay(x11->display);
+    g_free(x11->net_wm_name);
+    free(x11->randr.failed_conf);
     free(x11);
 }
 
@@ -254,8 +382,9 @@ static void vdagent_x11_set_clipboard_owner(struct vdagent_x11 *x11,
                           "ownership change, clearing");
                 once = 0;
             }
-            udscs_write(x11->vdagentd, VDAGENTD_CLIPBOARD_DATA, selection,
-                        VD_AGENT_CLIPBOARD_NONE, NULL, 0);
+            if (x11->vdagentd)
+                udscs_write(x11->vdagentd, VDAGENTD_CLIPBOARD_DATA, selection,
+                            VD_AGENT_CLIPBOARD_NONE, NULL, 0);
             if (curr_conv == x11->conversion_req) {
                 x11->conversion_req = next_conv;
                 x11->clipboard_data_size = 0;
@@ -272,7 +401,7 @@ static void vdagent_x11_set_clipboard_owner(struct vdagent_x11 *x11,
     if (new_owner == owner_none) {
         /* When going from owner_guest to owner_none we need to send a
            clipboard release message to the client */
-        if (x11->clipboard_owner[selection] == owner_guest) {
+        if (x11->clipboard_owner[selection] == owner_guest && x11->vdagentd) {
             udscs_write(x11->vdagentd, VDAGENTD_CLIPBOARD_RELEASE, selection,
                         0, NULL, 0);
         }
@@ -326,7 +455,7 @@ static int vdagent_x11_get_clipboard_selection(struct vdagent_x11 *x11,
 
 static void vdagent_x11_handle_event(struct vdagent_x11 *x11, XEvent event)
 {
-    int handled = 0;
+    int i, handled = 0;
     uint8_t selection;
 
     if (event.type == x11->xfixes_event_base) {
@@ -376,11 +505,14 @@ static void vdagent_x11_handle_event(struct vdagent_x11 *x11, XEvent event)
     switch (event.type) {
     case ConfigureNotify:
         // TODO: handle CrtcConfigureNotify, OutputConfigureNotify can be ignored.
-        if (event.xconfigure.window != x11->root_window)
+        for (i = 0; i < x11->screen_count; i++)
+            if (event.xconfigure.window == x11->root_window[i])
+                break;
+        if (i == x11->screen_count)
             break;
 
         handled = 1;
-        vdagent_x11_randr_handle_root_size_change(x11,
+        vdagent_x11_randr_handle_root_size_change(x11, i,
                 event.xconfigure.width, event.xconfigure.height);
         break;
     case MappingNotify:
@@ -824,7 +956,10 @@ static void vdagent_x11_send_selection_notify(struct vdagent_x11 *x11,
     res.xselection.selection = event->xselectionrequest.selection;
     res.xselection.target = event->xselectionrequest.target;
     res.xselection.time = event->xselectionrequest.time;
+
+    vdagent_x11_set_error_handler(x11, vdagent_x11_ignore_bad_window_handler);
     XSendEvent(x11->display, event->xselectionrequest.requestor, 0, 0, &res);
+    vdagent_x11_restore_error_handler(x11);
 
     if (!request) {
         vdagent_x11_next_selection_request(x11);
@@ -860,11 +995,16 @@ exit_loop:
     if (prop == None)
         prop = event->xselectionrequest.target;
 
+    vdagent_x11_set_error_handler(x11, vdagent_x11_ignore_bad_window_handler);
     XChangeProperty(x11->display, event->xselectionrequest.requestor, prop,
                     XA_ATOM, 32, PropModeReplace, (unsigned char *)&targets,
                     target_count);
-    vdagent_x11_print_targets(x11, selection, "sent", targets, target_count);
-    vdagent_x11_send_selection_notify(x11, prop, NULL);
+    if (vdagent_x11_restore_error_handler(x11) == 0) {
+        vdagent_x11_print_targets(x11, selection, "sent",
+                                  targets, target_count);
+        vdagent_x11_send_selection_notify(x11, prop, NULL);
+    } else
+        SELPRINTF("send_targets: Failed to sent, requestor window gone");
 }
 
 static void vdagent_x11_handle_selection_request(struct vdagent_x11 *x11)
@@ -937,11 +1077,17 @@ static void vdagent_x11_handle_property_delete_notify(struct vdagent_x11 *x11,
     } else {
         VSELPRINTF("Ending incr send of clipboard data");
     }
+    vdagent_x11_set_error_handler(x11, vdagent_x11_ignore_bad_window_handler);
     XChangeProperty(x11->display, sel_event->xselectionrequest.requestor,
                     x11->selection_req_atom,
                     sel_event->xselectionrequest.target, 8, PropModeReplace,
                     x11->selection_req_data + x11->selection_req_data_pos,
                     len);
+    if (vdagent_x11_restore_error_handler(x11)) {
+        SELPRINTF("incr sent failed, requestor window gone");
+        len = 0;
+    }
+
     x11->selection_req_data_pos += len;
 
     /* Note we must explictly send a 0 sized XChangeProperty to signal the
@@ -1091,21 +1237,33 @@ void vdagent_x11_clipboard_data(struct vdagent_x11 *x11, uint8_t selection,
     if (size > x11->max_prop_size) {
         unsigned long len = size;
         VSELPRINTF("Starting incr send of clipboard data");
-        x11->selection_req_data = data;
-        x11->selection_req_data_pos = 0;
-        x11->selection_req_data_size = size;
-        x11->selection_req_atom = prop;
+
+        vdagent_x11_set_error_handler(x11, vdagent_x11_ignore_bad_window_handler);
         XSelectInput(x11->display, event->xselectionrequest.requestor,
                      PropertyChangeMask);
         XChangeProperty(x11->display, event->xselectionrequest.requestor, prop,
                         x11->incr_atom, 32, PropModeReplace,
                         (unsigned char*)&len, 1);
-        vdagent_x11_send_selection_notify(x11, prop, x11->selection_req);
+        if (vdagent_x11_restore_error_handler(x11) == 0) {
+            x11->selection_req_data = data;
+            x11->selection_req_data_pos = 0;
+            x11->selection_req_data_size = size;
+            x11->selection_req_atom = prop;
+            vdagent_x11_send_selection_notify(x11, prop, x11->selection_req);
+        } else {
+            SELPRINTF("clipboard data sent failed, requestor window gone");
+            free(data);
+        }
     } else {
+        vdagent_x11_set_error_handler(x11, vdagent_x11_ignore_bad_window_handler);
         XChangeProperty(x11->display, event->xselectionrequest.requestor, prop,
                         event->xselectionrequest.target, 8, PropModeReplace,
                         data, size);
-        vdagent_x11_send_selection_notify(x11, prop, NULL);
+        if (vdagent_x11_restore_error_handler(x11) == 0)
+            vdagent_x11_send_selection_notify(x11, prop, NULL);
+        else
+            SELPRINTF("clipboard data sent failed, requestor window gone");
+
         free(data);
     }
 
@@ -1142,3 +1300,26 @@ void vdagent_x11_clipboard_release(struct vdagent_x11 *x11, uint8_t selection)
     /* Flush output buffers and consume any pending events */
     vdagent_x11_do_read(x11);
 }
+
+/* Function used to determine the default location to save file-xfers,
+   xdg desktop dir or xdg download dir. We error on the save side and use a
+   whitelist approach, so any unknown desktops will end up with saving
+   file-xfers to the xdg download dir, and opening the xdg download dir with
+   xdg-open when the file-xfer completes. */
+int vdagent_x11_has_icons_on_desktop(struct vdagent_x11 *x11)
+{
+    const char * const wms_with_icons_on_desktop[] = {
+        "Metacity", /* GNOME-2 or GNOME-3 fallback */
+        "Xfwm4",    /* XFCE */
+        "Marco",    /* Mate */
+        NULL
+    };
+    int i;
+
+    if (x11->net_wm_name)
+        for (i = 0; wms_with_icons_on_desktop[i]; i++)
+            if (!strcmp(x11->net_wm_name, wms_with_icons_on_desktop[i]))
+                return 1;
+
+    return 0;
+}
