@@ -1,6 +1,6 @@
 /*  vdagentd.c vdagentd (daemon) code
 
-    Copyright 2010-2012 Red Hat, Inc.
+    Copyright 2010-2013 Red Hat, Inc.
 
     Red Hat Authors:
     Hans de Goede <hdegoede@redhat.com>
@@ -34,6 +34,7 @@
 #include <sys/select.h>
 #include <sys/stat.h>
 #include <spice/vd_agent.h>
+#include <glib.h>
 
 #include "udscs.h"
 #include "vdagentd-proto.h"
@@ -58,22 +59,19 @@ static const char *uinput_device = "/dev/uinput";
 static int debug = 0;
 static struct udscs_server *server = NULL;
 static struct vdagent_virtio_port *virtio_port = NULL;
-#ifdef HAVE_SESSION_INFO
+static GHashTable *active_xfers = NULL;
 static struct session_info *session_info = NULL;
-#endif
 static struct vdagentd_uinput *uinput = NULL;
 static VDAgentMonitorsConfig *mon_config = NULL;
 static uint32_t *capabilities = NULL;
 static int capabilities_size = 0;
-#ifdef HAVE_SESSION_INFO
 static const char *active_session = NULL;
-#else
 static unsigned int session_count = 0;
-#endif
 static struct udscs_connection *active_session_conn = NULL;
 static int agent_owns_clipboard[256] = { 0, };
 static int quit = 0;
 static int retval = 0;
+static int client_connected = 0;
 
 /* utility functions */
 /* vdagentd <-> spice-client communication handling */
@@ -96,6 +94,7 @@ static void send_capabilities(struct vdagent_virtio_port *vport,
     VD_AGENT_SET_CAPABILITY(caps->caps, VD_AGENT_CAP_REPLY);
     VD_AGENT_SET_CAPABILITY(caps->caps, VD_AGENT_CAP_CLIPBOARD_BY_DEMAND);
     VD_AGENT_SET_CAPABILITY(caps->caps, VD_AGENT_CAP_CLIPBOARD_SELECTION);
+    VD_AGENT_SET_CAPABILITY(caps->caps, VD_AGENT_CAP_SPARSE_MONITORS_CONFIG);
 
     vdagent_virtio_port_write(vport, VDP_CLIENT_PORT,
                               VD_AGENT_ANNOUNCE_CAPABILITIES, 0,
@@ -103,6 +102,15 @@ static void send_capabilities(struct vdagent_virtio_port *vport,
     free(caps);
 }
 
+static void do_client_disconnect(void)
+{
+    if (client_connected) {
+        udscs_server_write_all(server, VDAGENTD_CLIENT_DISCONNECTED, 0, 0,
+                               NULL, 0);
+        client_connected = 0;
+    }
+}
+
 static void do_client_monitors(struct vdagent_virtio_port *vport, int port_nr,
     VDAgentMessage *message_header, VDAgentMonitorsConfig *new_monitors)
 {
@@ -159,8 +167,14 @@ static void do_client_capabilities(struct vdagent_virtio_port *vport,
         }
     }
     memcpy(capabilities, caps->caps, capabilities_size * sizeof(uint32_t));
-    if (caps->request)
+    if (caps->request) {
+        /* Report the previous client has disconneced. */
+        do_client_disconnect();
+        if (debug)
+            syslog(LOG_DEBUG, "New client connected");
+        client_connected = 1;
         send_capabilities(vport, 0);
+    }
 }
 
 static void do_client_clipboard(struct vdagent_virtio_port *vport,
@@ -215,6 +229,64 @@ static void do_client_clipboard(struct vdagent_virtio_port *vport,
                 data, size);
 }
 
+static void cancel_file_xfer(struct vdagent_virtio_port *vport,
+                             const char *msg, uint32_t id)
+{
+    VDAgentFileXferStatusMessage status = {
+        .id = id,
+        .result = VD_AGENT_FILE_XFER_STATUS_CANCELLED,
+    };
+    syslog(LOG_WARNING, msg, id);
+    if (vport)
+        vdagent_virtio_port_write(vport, VDP_CLIENT_PORT,
+                                  VD_AGENT_FILE_XFER_STATUS, 0,
+                                  (uint8_t *)&status, sizeof(status));
+}
+
+static void do_client_file_xfer(struct vdagent_virtio_port *vport,
+                                VDAgentMessage *message_header,
+                                uint8_t *data)
+{
+    uint32_t msg_type, id;
+    struct udscs_connection *conn;
+
+    switch (message_header->type) {
+    case VD_AGENT_FILE_XFER_START: {
+        VDAgentFileXferStartMessage *s = (VDAgentFileXferStartMessage *)data;
+        if (!active_session_conn) {
+            cancel_file_xfer(vport,
+               "Could not find an agent connnection belonging to the "
+               "active session, cancelling client file-xfer request %u",
+               s->id);
+            return;
+        }
+        udscs_write(active_session_conn, VDAGENTD_FILE_XFER_START, 0, 0,
+                    data, message_header->size);
+        return;
+    }
+    case VD_AGENT_FILE_XFER_STATUS: {
+        VDAgentFileXferStatusMessage *s = (VDAgentFileXferStatusMessage *)data;
+        msg_type = VDAGENTD_FILE_XFER_STATUS;
+        id = s->id;
+        break;
+    }
+    case VD_AGENT_FILE_XFER_DATA: {
+        VDAgentFileXferDataMessage *d = (VDAgentFileXferDataMessage *)data;
+        msg_type = VDAGENTD_FILE_XFER_DATA;
+        id = d->id;
+        break;
+    }
+    }
+
+    conn = g_hash_table_lookup(active_xfers, GUINT_TO_POINTER(id));
+    if (!conn) {
+        if (debug)
+            syslog(LOG_DEBUG, "Could not find file-xfer %u (cancelled?)", id);
+        return;
+    }
+    udscs_write(conn, msg_type, 0, 0, data, message_header->size);
+}
+
 int virtio_port_read_complete(
         struct vdagent_virtio_port *vport,
         int port_nr,
@@ -284,6 +356,15 @@ int virtio_port_read_complete(
         }
         do_client_clipboard(vport, message_header, data);
         break;
+    case VD_AGENT_FILE_XFER_START:
+    case VD_AGENT_FILE_XFER_STATUS:
+    case VD_AGENT_FILE_XFER_DATA:
+        do_client_file_xfer(vport, message_header, data);
+        break;
+    case VD_AGENT_CLIENT_DISCONNECTED:
+        vdagent_virtio_port_reset(vport, VDP_CLIENT_PORT);
+        do_client_disconnect();
+        break;
     default:
         syslog(LOG_WARNING, "unknown message type %d, ignoring",
                message_header->type);
@@ -444,7 +525,6 @@ static void check_xorg_resolution(void)
     }
 }
 
-#ifdef HAVE_SESSION_INFO
 static int connection_matches_active_session(struct udscs_connection **connp,
     void *priv)
 {
@@ -460,7 +540,6 @@ static int connection_matches_active_session(struct udscs_connection **connp,
     *conn_ret = *connp;
     return 1;
 }
-#endif
 
 void release_clipboards(void)
 {
@@ -475,18 +554,27 @@ void release_clipboards(void)
     }
 }
 
-void update_active_session_connection(void)
+void update_active_session_connection(struct udscs_connection *new_conn)
 {
-#ifdef HAVE_SESSION_INFO
-    int n;
-    struct udscs_connection *new_conn = NULL;
-    if (!active_session)
-        active_session = session_info_get_active_session(session_info);
-
-    n = udscs_server_for_all_clients(server, connection_matches_active_session,
-                                     (void*)&new_conn);
-    if (n != 1)
+    if (session_info) {
         new_conn = NULL;
+        if (!active_session)
+            active_session = session_info_get_active_session(session_info);
+        session_count = udscs_server_for_all_clients(server,
+                                         connection_matches_active_session,
+                                         (void*)&new_conn);
+    } else {
+        if (new_conn)
+            session_count++;
+        else
+            session_count--;
+    }
+
+    if (new_conn && session_count != 1) {
+        syslog(LOG_ERR, "multiple agents in one session, "
+               "disabling agent to avoid potential information leak");
+        new_conn = NULL;
+    }
 
     if (new_conn == active_session_conn)
         return;
@@ -498,18 +586,24 @@ void update_active_session_connection(void)
         udscs_write(active_session_conn, VDAGENTD_MONITORS_CONFIG, 0, 0,
                     (uint8_t *)mon_config, sizeof(VDAgentMonitorsConfig) +
                     mon_config->num_of_monitors * sizeof(VDAgentMonConfig));
-#endif
 
     release_clipboards();
 
     check_xorg_resolution();    
 }
 
+gboolean remove_active_xfers(gpointer key, gpointer value, gpointer conn)
+{
+    if (value == conn) {
+        cancel_file_xfer(virtio_port, "Agent disc; cancelling file-xfer %u",
+                         GPOINTER_TO_UINT(key));
+        return 1;
+    } else
+        return 0;
+}
+
 void agent_connect(struct udscs_connection *conn)
 {
-#ifdef HAVE_SESSION_INFO
-    uint32_t pid;
-#endif
     struct agent_data *agent_data;
 
     agent_data = calloc(1, sizeof(*agent_data));
@@ -518,47 +612,29 @@ void agent_connect(struct udscs_connection *conn)
         udscs_destroy_connection(&conn);
         return;
     }
-#ifdef HAVE_SESSION_INFO
-    pid = udscs_get_peer_cred(conn).pid;
-    agent_data->session = session_info_session_for_pid(session_info, pid);
-#else
-    session_count++;
-    if (session_count == 1) {
-        active_session_conn = conn;
-    } else {
-        /* disable communication with agents when we've got multiple
-         * connections to the vdagentd and no consolekit since we can't
-         * know to which one we should send data
-         */
-        syslog(LOG_ERR, "Trying to use multiple vdagent without ConsoleKit, "
-               "disabling vdagent to avoid potential information leak");
-        active_session_conn = NULL;
+
+    if (session_info) {
+        uint32_t pid = udscs_get_peer_cred(conn).pid;
+        agent_data->session = session_info_session_for_pid(session_info, pid);
     }
-#endif
 
     udscs_set_user_data(conn, (void *)agent_data);
     udscs_write(conn, VDAGENTD_VERSION, 0, 0,
                 (uint8_t *)VERSION, strlen(VERSION) + 1);
-    update_active_session_connection();
+    update_active_session_connection(conn);
 }
 
 void agent_disconnect(struct udscs_connection *conn)
 {
     struct agent_data *agent_data = udscs_get_user_data(conn);
 
-#ifndef HAVE_SESSION_INFO
-    if (conn == active_session_conn)
-        active_session_conn = NULL;
-#endif
+    g_hash_table_foreach_remove(active_xfers, remove_active_xfers, conn);
 
     free(agent_data->session);
     agent_data->session = NULL;
-    update_active_session_connection();
+    update_active_session_connection(NULL);
 
     free(agent_data);
-#ifndef HAVE_SESSION_INFO
-    session_count--;
-#endif
 }
 
 void agent_read_complete(struct udscs_connection **connp,
@@ -614,6 +690,21 @@ void agent_read_complete(struct udscs_connection **connp,
             return;
         }
         break;
+    case VDAGENTD_FILE_XFER_STATUS:{
+        VDAgentFileXferStatusMessage status;
+        status.id = header->arg1;
+        status.result = header->arg2;
+        vdagent_virtio_port_write(virtio_port, VDP_CLIENT_PORT,
+                                  VD_AGENT_FILE_XFER_STATUS, 0,
+                                  (uint8_t *)&status, sizeof(status));
+        if (status.result == VD_AGENT_FILE_XFER_STATUS_CAN_SEND_DATA)
+            g_hash_table_insert(active_xfers, GUINT_TO_POINTER(status.id),
+                                *connp);
+        else
+            g_hash_table_remove(active_xfers, GUINT_TO_POINTER(status.id));
+        break;
+    }
+
     default:
         syslog(LOG_ERR, "unknown message from vdagent: %u, ignoring",
                header->type);
@@ -626,14 +717,21 @@ void agent_read_complete(struct udscs_connection **connp,
 static void usage(FILE *fp)
 {
     fprintf(fp,
-            "vdagentd\n"
-            "options:\n"
+            "Usage: spice-vdagentd [OPTIONS]\n\n"
+            "Spice guest agent daemon.\n\n"
+            "Options:\n"
             "  -h         print this text\n"
             "  -d         log debug messages (use twice for extra info)\n"
             "  -s <port>  set virtio serial port  [%s]\n"
             "  -u <dev>   set uinput device       [%s]\n"
-            "  -x         don't daemonize\n",
-            portdev, uinput_device);
+            "  -x         don't daemonize\n"
+#ifdef HAVE_CONSOLE_KIT
+            "  -X         Disable console kit integration\n"
+#endif
+#ifdef HAVE_LIBSYSTEMD_LOGIN
+            "  -X         Disable systemd-logind integration\n"
+#endif
+            ,portdev, uinput_device);
 }
 
 void daemonize(void)
@@ -666,9 +764,7 @@ void main_loop(void)
 {
     fd_set readfds, writefds;
     int n, nfds;
-#ifdef HAVE_SESSION_INFO
     int ck_fd = 0;
-#endif
 
     while (!quit) {
         FD_ZERO(&readfds);
@@ -679,12 +775,12 @@ void main_loop(void)
         if (n >= nfds)
             nfds = n + 1;
 
-#ifdef HAVE_SESSION_INFO
-        ck_fd = session_info_get_fd(session_info);
-        FD_SET(ck_fd, &readfds);
-        if (ck_fd >= nfds)
-            nfds = ck_fd + 1;
-#endif
+        if (session_info) {
+            ck_fd = session_info_get_fd(session_info);
+            FD_SET(ck_fd, &readfds);
+            if (ck_fd >= nfds)
+                nfds = ck_fd + 1;
+        }
 
         n = select(nfds, &readfds, &writefds, NULL, NULL);
         if (n == -1) {
@@ -700,25 +796,27 @@ void main_loop(void)
         if (virtio_port) {
             vdagent_virtio_port_handle_fds(&virtio_port, &readfds, &writefds);
             if (!virtio_port) {
+                int old_client_connected = client_connected;
                 syslog(LOG_CRIT,
                        "AIIEEE lost spice client connection, reconnecting");
                 virtio_port = vdagent_virtio_port_create(portdev,
                                                      virtio_port_read_complete,
                                                      NULL);
-            }
-            if (!virtio_port) {
-                syslog(LOG_CRIT, "Fatal error opening vdagent virtio channel");
-                retval = 1;
-                break;
+                if (!virtio_port) {
+                    syslog(LOG_CRIT,
+                           "Fatal error opening vdagent virtio channel");
+                    retval = 1;
+                    break;
+                }
+                do_client_disconnect();
+                client_connected = old_client_connected;
             }
         }
 
-#ifdef HAVE_SESSION_INFO
-        if (FD_ISSET(ck_fd, &readfds)) {
+        if (session_info && FD_ISSET(ck_fd, &readfds)) {
             active_session = session_info_get_active_session(session_info);
-            update_active_session_connection();
+            update_active_session_connection(NULL);
         }
-#endif
     }
 }
 
@@ -731,10 +829,11 @@ int main(int argc, char *argv[])
 {
     int c;
     int do_daemonize = 1;
+    int want_session_info = 1;
     struct sigaction act;
 
     for (;;) {
-        if (-1 == (c = getopt(argc, argv, "-dhxs:u:")))
+        if (-1 == (c = getopt(argc, argv, "-dhxXs:u:")))
             break;
         switch (c) {
         case 'd':
@@ -749,10 +848,14 @@ int main(int argc, char *argv[])
         case 'x':
             do_daemonize = 0;
             break;
+        case 'X':
+            want_session_info = 0;
+            break;
         case 'h':
             usage(stdout);
             return 0;
         default:
+            fputs("\n", stderr);
             usage(stderr);
             return 1;
         }
@@ -797,16 +900,12 @@ int main(int argc, char *argv[])
     }
 #endif
 
-#ifdef HAVE_SESSION_INFO
-    session_info = session_info_create(debug);
-    if (!session_info) {
-        syslog(LOG_CRIT, "Fatal could not get session information");
-        vdagentd_uinput_destroy(&uinput);
-        udscs_destroy_server(server);
-        return 1;
-    }
-#endif
+    if (want_session_info)
+        session_info = session_info_create(debug);
+    if (!session_info)
+        syslog(LOG_WARNING, "no session info, max 1 session agent allowed");
 
+    active_xfers = g_hash_table_new(g_direct_hash, g_direct_equal);
     main_loop();
 
     release_clipboards();
@@ -814,9 +913,7 @@ int main(int argc, char *argv[])
     vdagentd_uinput_destroy(&uinput);
     vdagent_virtio_port_flush(&virtio_port);
     vdagent_virtio_port_destroy(&virtio_port);
-#ifdef HAVE_SESSION_INFO
     session_info_destroy(session_info);
-#endif
     udscs_destroy_server(server);
     if (unlink(VDAGENTD_SOCKET) != 0)
         syslog(LOG_ERR, "unlink %s: %s", VDAGENTD_SOCKET, strerror(errno));
